﻿
/******************************************************************
 * ==============================================================
 *					Map File Format
 * ==============================================================
 *
 * int8_t magic[4] = 'CYMF' (Colysis Map File) : 4 bytes
 * uint8_t version[3] : major, minor, patch
 *
 * uint8_t flags
 * if (flags):
 *   case MapFlags::CustomTilemap:
 *     uint32_t filepath_buffer_size
 *     char filepath_buffer[filepath_buffer_size]
 *     uint32_t tile_size[2] : width, height
 *     uint32_t tile_count[2] : width, height
 *
 * uint32_t map_size[3] : x, y, specials
 * int32_t tiles[x * y + specials * (sizeof(MapSpecial)/4)] : tilemap data
 * // a special tile has a type and its tilemap data
 * if tiles[i] == Special:
 *   uint8_t type
 *   MapSpecial data
 * else:
 *   uint8_t tile
 *
 * uint32_t player_spawn[2] : x, y
 * uint32_t number_enemy_spawns
 * uint32_t enemy_spawns[number_enemy_spawns][2] : x, y
 *
*******************************************************************/

#pragma once
#include <SFML/Graphics.hpp>

#define STR(name) #name
#define STR_VALUE(name) STR(name)

#define MAP_VERSION_MAJOR 0
#define MAP_VERSION_MINOR 1
#define MAP_VERSION_PATCH 0
#define MAP_VERSION (MAP_VERSION_MAJOR << 16 | MAP_VERSION_MINOR << 8 | MAP_VERSION_PATCH)
#define MAP_VERSION_STRING STR_VALUE(MAP_VERSION_MAJOR) "." STR_VALUE(MAP_VERSION_MINOR) "." STR_VALUE(MAP_VERSION_PATCH)


struct MapSpecial
{
	int type;
	uint32_t portal_end[2];
};

enum class MapFlags : uint8_t
{
	None = 0,
	CustomTilemap = 1 << 0,
	UNDEFINED = 1 << 1,
};

enum class TileType : int
{
	Wall = -1,
	Special = 0,
	Other = 1,
};


struct MapError
{
	std::string message;
	int code;
};


class Map : public sf::Drawable, public sf::Transformable
{
	static_assert(sizeof(MapSpecial)%4 == 0);

	struct {
		sf::Vector2u tile_size;
		sf::Vector2u tile_count;

		uint32_t wall_tiles;
		uint32_t other_tiles;
		uint32_t special_tiles;

		sf::Texture texture;
		sf::VertexArray vertices;
	} atlas;

	sf::Vector3<size_t> map_size; // width * height, width, height
	int *tiles = nullptr;

	std::vector<std::pair<sf::Vector2u, MapSpecial>> specials;

	std::vector<sf::Vector2u> enemy_spawns;
	sf::Vector2u player_spawn;

	MapError populate_vertex_array();

	void LoadDefaultTilemap();

public:
	Map();

	MapError load(const std::filesystem::path& path);

	sf::Vector2u getSize() const;

	void draw(sf::RenderTarget& target, sf::RenderStates states) const override;

#ifdef _DEBUG
	void debug();
#endif
};


